/*
 * stalk/bi_func.c
 * Copyright (c) 1993 Gerald Weiler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * crypt, index, rindex, random, and strsub taken from LambdaMOO,
 * by Pavel Curtis
 */

#include "include.h"
#include "bi_func.h"
#include "stalk.h"
#include "buffer.h"
#include "stuff.h"
#include "task.h"
#include "parser.h"
#include "var.h"
#include "interpret.h"
#include "functions.h"

void bi_crypt(t, args)
     task *t;
     var *args;
{
  char salt[3];
  static char saltstuff[] =
    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";
  extern char *crypt P((const char *, const char *));
  static var_type types[] = { V_STRING, V_STRING };
  Error e;

  if ((e = inv_args(args, 1, 2, types)) != E_NONE) {
    in_push(t, make_error(t, e));
    return;
  }

  if (args->num == 1) {
    salt[0] = saltstuff[rand() % strlen(saltstuff)];
    salt[1] = saltstuff[rand() % strlen(saltstuff)];
  } else if (!(salt[0] = args->list[1]->string[0])
	     || !(salt[1] = args->list[1]->string[1])) {
    in_push(t, make_error(t, E_INVARG));
    return;
  }
  salt[2] = '\0';
  in_push(t, newstring(mystrdup(crypt(args->list[0]->string, salt))));
}

void bi_index(t, args)
     task *t;
     var *args;
{
  int case_matters;
  static var_type types[] = { V_STRING, V_STRING, V_NUM };
  Error e;

  if ((e = inv_args(args, 2, 3, types)) != E_NONE) {
    in_push(t, make_error(t, e));
    return;
  }

  if (args->num == 3)
    case_matters = args->list[2]->num;
  else
    case_matters = 0;

  in_push(t, newnum(strindex(args->list[0]->string, args->list[1]->string,
			     case_matters)));
}

void bi_rindex(t, args)
     task *t;
     var *args;
{
  int case_matters;
  static var_type types[] = { V_STRING, V_STRING, V_NUM };
  Error e;

  if ((e = inv_args(args, 2, 3, types)) != E_NONE) {
    in_push(t, make_error(t, e));
    return;
  }

  if (args->num == 3)
    case_matters = args->list[2]->num;
  else
    case_matters = 0;

  in_push(t, newnum(strrindex(args->list[0]->string, args->list[1]->string,
			      case_matters)));
}

void bi_random(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_NUM };
  Error e;

  if ((e = inv_args(args, 1, 1, types)) != E_NONE) {
    in_push(t, make_error(t, e));
    return;
  }

  if (args->list[0]->num <= 1) {
    in_push(t, make_error(t, E_INVARG));
    return;
  }

  in_push(t, newnum(rand() % args->list[0]->num + 1));
}

void bi_strsub(t, args)
     task *t;
     var *args;
{
  int case_matters;
  static var_type types[] = { V_STRING, V_STRING, V_STRING, V_NUM };
  Error e;

  if ((e = inv_args(args, 3, 4, types)) != E_NONE) {
    in_push(t, make_error(t, e));
    return;
  }

  if (args->num == 4)
    case_matters = args->list[3]->num;
  else
    case_matters = 0;

  in_push(t, newstring(strsub(args->list[0]->string, args->list[1]->string,
			      args->list[2]->string, case_matters)));
}

/*
 * bi_readdir returns a list of strings, where each element is one of
 * the elements of the directory that was requested.
 */
void bi_readdir(t, args)
     task *t;
     var *args;
{
  var *r;
  int x, y;
  DIR *dir;
  struct dirent *dir_entry;
  static var_type types[] = { V_STRING };
  Error e;
  char *filename;

  if ((e = inv_args(args, 1, 1, types)) != E_NONE) {
    in_push(t, make_error(t, e));
    return;
  }

  /*
   * this assumes the directory won't change as it's being read.
   * if things are added between when the count is done and when
   * the actual read is done, a file may be missing from the directory.
   * (not really a problem...)
   * if a file is deleted, then r->num will just have to be changed
   * to reflect this...
   * if the directory's deleted, well then... you're SOL.
   */
  filename = expand_filename(args->list[0]->string);

  if ((dir = opendir(filename))) {
    for (x=0; readdir(dir); x++)
      ;
    r = newlist(x);
    closedir(dir);
    dir = opendir(filename);
    for (y=0; y<x; y++) {
      if (!(dir_entry = readdir(dir))) {
	r->num = y;
	break;
      } else {
	r->list[y] = newstring(mystrdup(dir_entry->d_name));
      }
    }
    closedir(dir);
  } else
    r = newnum(0);

  in_push(t, r);

  if (filename)
    free(filename);
}

/*
 * bi_stat returns data in the form
 * {mode, uid/user, gid/group, size, atime, mtime, ctime}
 * whether uid/user and gid/group are numbers or string depends upon
 * the compile-time options. mode will be a 10 character string ala ls -l.
 * all other values are numbers.
 * the time values are in a form that could be put through ctime()
 */
void bi_stat(t, args)
     task *t;
     var *args;
{
  var *r;
  struct stat buf;
#ifndef NO_PW
  struct passwd *file_user;
#endif
#ifndef NO_GR
  struct group *file_group;
#endif
  static var_type types[] = { V_STRING };
  Error e;
  char *filename;

  if ((e = inv_args(args, 1, 1, types)) != E_NONE) {
    in_push(t, make_error(t, e));
    return;
  }

  filename = expand_filename(args->list[0]->string);
  if (!stat(filename, &buf)) {
    r = newlist(7);

    r->list[0] = newstring(mystrdup("----------"));
    if (S_ISDIR(buf.st_mode))
      r->list[0]->string[0] = 'd';
    else if (S_ISCHR(buf.st_mode))
      r->list[0]->string[0] = 'c';
    else if (S_ISBLK(buf.st_mode))
      r->list[0]->string[0] = 'b';
    else if (S_ISFIFO(buf.st_mode))
      r->list[0]->string[0] = 'p';
#ifdef S_ISLNK
    else if (S_ISLNK(buf.st_mode))
      r->list[0]->string[0] = 'l';
#endif /* S_ISLNK */
#ifdef S_ISSOCK
    else if (S_ISSOCK(buf.st_mode))
      r->list[0]->string[0] = 's';
#endif /* S_ISSOCK */
    r->list[0]->string[1] = (buf.st_mode & S_IRUSR) ? 'r' : '-';
    r->list[0]->string[2] = (buf.st_mode & S_IWUSR) ? 'w' : '-';
    r->list[0]->string[3] = (buf.st_mode & S_IXUSR)
      ? ((buf.st_mode & S_ISUID) ? 's' : 'x')
	: ((buf.st_mode & S_ISUID) ? 'S' : '-');
    r->list[0]->string[4] = (buf.st_mode & S_IRGRP) ? 'r' : '-';
    r->list[0]->string[5] = (buf.st_mode & S_IWGRP) ? 'w' : '-';
    r->list[0]->string[6] = (buf.st_mode & S_IXGRP)
      ? ((buf.st_mode & S_ISGID) ? 's' : 'x')
	: ((buf.st_mode & S_ISGID) ? 'S' : '-');
    r->list[0]->string[7] = (buf.st_mode & S_IROTH) ? 'r' : '-';
    r->list[0]->string[8] = (buf.st_mode & S_IWOTH) ? 'w' : '-';
#ifdef S_ISVTX
    r->list[0]->string[9] = (buf.st_mode & S_IXOTH)
      ? ((buf.st_mode & S_ISVTX) ? 't' : 'x')
	: ((buf.st_mode & S_ISVTX) ? 'T' : '-');
#else
    r->list[0]->string[9] = (buf.st_mode & S_IXOTH) ? 'x' : '-';
#endif /* S_ISVTX */

#ifndef NO_PW
    if ((file_user = getpwuid(buf.st_uid)))
      r->list[1] = newstring(mystrdup(file_user->pw_name));
    else
#endif /* NO_PW */
      r->list[1] = newnum((int)buf.st_uid);

#ifndef NO_GR
    if ((file_group = getgrgid(buf.st_gid)))
      r->list[2] = newstring(mystrdup(file_group->gr_name));
    else
#endif /* NO_GR */
      r->list[2] = newnum((int)buf.st_gid);

    r->list[3] = newnum((int)buf.st_size);
    r->list[4] = newnum((int)buf.st_atime);
    r->list[5] = newnum((int)buf.st_mtime);
    r->list[6] = newnum((int)buf.st_ctime);
    in_push(t, r);
  } else
    in_push(t, newnum(0));

  if (filename)
    free(filename);
}

void bi_eval(t, args)
     task *t;
     var *args;
{
  char *s;
  statement *temp;

  s = vartostr(args);
  temp = my_parse(t->buf, s);
  free(s);
  if (temp) {
    program *p;

    p = NEW(program);
    p->name = ("* evaled program *");
    p->ref = 0;
    p->evaled = 1;
    p->debug = 1;
    p->base = temp;
    exec_task(t, p, NULL, 0);
  } else
    in_push(t, NULL);
}

void bi_length(t, args)
     task *t;
     var *args;
{
  if ((args->type != V_LIST) || (args->num != 1)) {
    in_push(t, make_error(t, E_ARGS));
    return;
  }

  if (args->list[0]->type == V_STRING)
    in_push(t, newnum((int)strlen(args->list[0]->string)));
  else if (args->list[0]->type == V_LIST)
    in_push(t, newnum(args->list[0]->num));
  else
    in_push(t, make_error(t, E_TYPE));
}

/*
 * time functions taken from LambdaMOO, by Pavel Curtis
 */
void bi_time(t, args)
     task *t;
     var *args;
{
  if ((args->type != V_LIST) || args->num)
    in_push(t, make_error(t, E_ARGS));
  else
    in_push(t, newnum(time(0)));
}

void bi_ctime(t, args)
     task *t;
     var *args;
{
  var *r;
  time_t c;
  char string[50];
  static var_type types[] = { V_NUM };
  Error e;

  if ((e = inv_args(args, 0, 1, types)) != E_NONE) {
    in_push(t, make_error(t, e));
    return;
  }

  c = args->num ? args->list[0]->num : time(0);

#ifndef NO_STRFTIME
  strftime(string, 50, "%a %b %d %H:%M:%S %Y %Z", localtime(&c));
  if (string[8] == '0')
    string[8] = ' ';
  r = newstring(mystrdup(string));
#else
  r = newstring(mystrdup(ctime(&c)));
  if (r->string[8] == '0')
    r->string[8] = ' ';
#endif /* NO_STRFTIME */
  in_push(t, r);
}

void bi_tostr(t, args)
     task *t;
     var *args;
{
  in_push(t, newstring(vartostr(args)));
}

void bi_tonum(t, args)
     task *t;
     var *args;
{
  int x;

  if ((args->type != V_LIST) || (args->num != 1))
    in_push(t, make_error(t, E_ARGS));
  else
    switch (args->list[0]->type) {
    case V_NUM:
      in_push(t, newnum(args->list[0]->num));
      break;
    case V_ERR:
      in_push(t, newnum((int)args->list[0]->err));
      break;
    case V_STRING:
      (void) sscanf(args->list[0]->string, "%d", &x);
      in_push(t, newnum(x));
      break;
    case V_LIST:
    default:
      in_push(t, make_error(t, E_TYPE));
      break;
    }
}

void bi_typeof(t, args)
     task *t;
     var *args;
{
  if ((args->type != V_LIST) || (args->num != 1))
    in_push(t, make_error(t, E_ARGS));
  else
    in_push(t, newnum(args->list[0]->type));
}

void bi_vars(t, args)
     task *t;
     var *args;
{
  if (!args || (args->type != V_LIST) || args->num)
    in_push(t, make_error(t, E_ARGS));
  else
    in_push(t, varlist(t));
}

void bi_split_buffer(t, args)
     task *t;
     var *args;
{
  if (args->num > 1)
    in_push(t, make_error(t, E_ARGS));
  else if (args->num && (args->list[0]->type != V_NUM))
    in_push(t, make_error(t, E_TYPE));
  else
    in_push(t, newnum((new_buffer)(t->buf, "* new buffer *",
				   args->num ? args->list[0]->num: 0)));
}

void bi_send_event(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_NUM, V_STRING };
  int target;
  Error e;

  if ((e = inv_args(args, 2, 2, types)) != E_NONE) {
    in_push(t, make_error(t, e));
    return;
  }

  target = args->list[0]->num;
  if ((target<1) || (target>num_buffers) || !get_buffer(target-1))
    in_push(t, make_error(t, E_INVARG));
  else {
    send_buffer_event(get_buffer(target-1), args->list[1]->string, 1);
    in_push(t, NULL);
  }
}

void bi_printbuffer(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_NUM, V_STRING };
  int target;
  Error e;

  if ((e = inv_args(args, 2, 2, types)) != E_NONE) {
    in_push(t, make_error(t, e));
    return;
  }

  target = args->list[0]->num;
  if ((target<1) || (target>num_buffers) || !get_buffer(target-1))
    in_push(t, make_error(t, E_INVARG));
  else {
    send_buffer_event(get_buffer(target-1), args->list[1]->string, 0);
    in_push(t, NULL);
  }
}

void bi_title(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_STRING };
  Error e;

  if ((e = inv_args(args, 1, 1, types)) != E_NONE) {
    in_push(t, make_error(t, e));
    return;
  }

  if (t->buf->name)
    free(t->buf->name);
  t->buf->name = mystrdup(args->list[0]->string);
  (print_title)(t->buf);
  in_push(t, NULL);
}

void bi_getenv(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_STRING };
  Error e;
  char *s;

  if ((e = inv_args(args, 1, 1, types)) != E_NONE)
    in_push(t, make_error(t, e));
  else if ((s = getenv(args->list[0]->string)))
    in_push(t, newstring(mystrdup(s)));
  else
    in_push(t, newnum(0));
}

void bi_char(t, args)
     task *t;
     var *args;
{
  int n;
  unsigned char s[2];

  if (args->num != 1) {
    in_push(t, make_error(t, E_ARGS));
    return;
  } else if ((args->list[0]->type != V_NUM) &&
	     (args->list[0]->type != V_STRING)) {
    in_push(t, make_error(t, E_TYPE));
    return;
  }

  if (args->list[0]->type == V_NUM) {
    n = args->list[0]->num;
    if ((n < 0) || (n > 255))
      in_push(t, make_error(t, E_INVARG));
    else {
      s[0] = n;
      s[1] = '\0';
      in_push(t, newstring(mystrdup((char *) s)));
    }
  } else {
    if (strlen(args->list[0]->string) != 1)
      in_push(t, make_error(t, E_INVARG));
    else {
      s[0] = args->list[0]->string[0];
      in_push(t, newnum((int) s[0]));
    }
  }
}

void bi_get_input(t, args)
     task *t;
     var *args;
{
  var *v;

  if (args && args->num) {
    in_push(t, make_error(t, E_ARGS));
    return;
  }

  t->buf->keyboard.line[t->buf->keyboard.len] = '\0';
  v = newlist(2);
  v->list[0] = newstring(mystrdup(t->buf->keyboard.line));
  v->list[1] = newnum(t->buf->keyboard.pos);
  in_push(t, v);
}

void bi_set_input(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_STRING, V_NUM };
  Error e;

  if ((e = inv_args(args, 2, 2, types)) != E_NONE)
    in_push(t, make_error(t, e));
  else {
    strcpy(t->buf->keyboard.line, args->list[0]->string);
    t->buf->keyboard.len = strlen(t->buf->keyboard.line);
    t->buf->keyboard.pos = args->list[1]->num;
    (redraw)(t->buf);
    in_push(t, NULL);
  }
}

void bi_quit(t, args)
     task *t;
     var *args;
{
  if (args && args->num) {
    in_push(t, make_error(t, E_ARGS));
    return;
  }

  done = 1;
  in_push(t, NULL);
}

void bi_version(t, args)
     task *t;
     var *args;
{
  var *v;

  if (args && args->num) {
    in_push(t, make_error(t, E_ARGS));
    return;
  }

  v = newlist(1);
  v->list[0] = newstring(mystrdup(VERSION));
  in_push(t, v);
}

void bi_buffer_id(t, args)
     task *t;
     var *args;
{
  if (args && args->num) {
    in_push(t, make_error(t, E_ARGS));
    return;
  }

  in_push(t, newnum(t->buf->buf_id));
}
