/*
 * stalk/buffer.c
 * Copyright (c) 1993 Gerald Weiler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "include.h"
#include "buffer.h"
#include "lines.h"
#include "functions.h"
#include "stream.h"

buffer **buffers;
int num_buffers;
int num_buffer_slots;
int current_buffer;

buffer *newbuffer()
{
  buffer *b;

  b = NEW(buffer);

  b->buf_id = num_buffers+1;
  b->lines = init_line_queue();
  b->history = init_line_queue();
  init_streams(b);
  b->history_pos = -1;
  b->keyboard.len = 0;
  b->keyboard.pos = 0;
  b->keyboard.line[0] = '\0';
  b->local_bindings = NULL;
  if (num_buffers == num_buffer_slots)
    buffers = (buffer **)realloc(buffers,
                                 (num_buffer_slots += 5)*sizeof(buffer *));
  buffers[num_buffers++] = b;
  return b;
}

buffer *get_buffer(x)
     int x;
{
  return buffers[x];
}

void free_buffer(b)
     buffer *b;
{
}

void init_buffers()
{
  buffers = (buffer **)malloc(5*sizeof(buffer *));
  num_buffer_slots = 5;
  current_buffer = 0;
  num_buffers = 0;
}

void next_buffer(buf)
     buffer *buf;
{
  if (num_buffers == 1)
    return;

  (redraw)(buffers[current_buffer]);
  while(!buffers[current_buffer=(current_buffer+1)%num_buffers]);
  (redraw)(buffers[current_buffer]);
}

void split_buffer(buf)
     buffer *buf;
{
  (void)(new_buffer)(buf, "* new buffer *", 0);
}

void refresh(buf)
     buffer *buf;
{
  int x;

  for (x=0; x<num_buffers; x++)
    if (buffers[x]) {
      (redraw)(buffers[x]);
    }
}
