/*
 * stalk/debug.c
 * Copyright (c) 1993 Gerald Weiler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "include.h"
#include "debug.h"
#include "interpret.h"
#include "var.h"
#include "functions.h"
#include "stuff.h"
#include "task.h"

void init_debug(t)
     task *t;
{
  int x;

  t->error_pending = 0;
  t->error = 0;
  t->debug_task = 0;
  t->ignore_errors = 0;

  for (x=0; x<NUM_ERRORS; x++) {
    t->debug[x] = 1;
    t->debug_func[x] = NULL;
  }
}

void raise_error(t, err)
     task *t;
     Error err;
{
  if (!t || t->ignore_errors)
    return;

  /* we want to give the first error that and operation encountered */
  if (t->error_pending)
    return;

  if (!t->debug[(int) err])
    return;

  t->error_pending = 1;
  t->error = err;
}

int task_recursive(t, name)
     task *t;
     char *name;
{
  while(t) {
    if (!strcmp(name, t->program->name))
      return 1;
    t = t->parent;
  }

  return 0;
}

var *debug_func(t)
     task *t;
{
  int x;
  var *v1, *v2;

  v1 = newlist(5);

  v1->list[0] = newstring(mystrdup(t->program->name));
  v1->list[1] = var_ref(t->line_info->list[0]);
  v1->list[2] = var_ref(t->line_info->list[1]);

  v2 = newlist(t->line_info->list[2]->num);
  for (x=0; x<t->line_info->list[2]->num; x++) {
    v2->list[x] = newlist(2);
    v2->list[x]->list[0] = var_ref(t->line_info->list[2]->list[x]);
    v2->list[x]->list[1] = get_var(t->local,
				   t->line_info->list[2]->list[x]->string);
  }
  v1->list[3] = v2;

  v2 = newlist(t->line_info->list[2]->num);
  for (x=0; x<t->line_info->list[2]->num; x++) {
    v2->list[x] = newlist(2);
    v2->list[x]->list[0] = var_ref(t->line_info->list[2]->list[x]);
    v2->list[x]->list[1] = get_var(t->buf->global,
				   t->line_info->list[2]->list[x]->string);
  }
  v1->list[4] = v2;

  return v1;
}

void debug_dump_func(t)
     task *t;
{
  int x;
  var *v;
  char debug_line[1000];
  char *s;

  (displaystr)(t->buf, t->line_info->list[1]->string);

  /* print variables */
  for (x=0; x<t->line_info->list[2]->num; x++) {
    v = get_var(t->local, t->line_info->list[2]->list[x]->string);
    sprintf(debug_line, "%s: %s", t->line_info->list[2]->list[x]->string,
	    s=vartostr(v));
    free(s);
    var_free(v);
    (displaystr)(t->buf, debug_line);
  }
  for (x=0; x<t->line_info->list[3]->num; x++) {
    v = get_var(t->buf->global,
		t->line_info->list[3]->list[x]->string);
    sprintf(debug_line, "$%s: %s", t->line_info->list[3]->list[x]->string,
	    s=vartostr(v));
    free(s);
    var_free(v);
    (displaystr)(t->buf, debug_line);
  }
}

void debug_task(this_task, err)
     task *this_task;
     Error err;
{
  char debug_line[1000];
  task *t;
  int x;
  var *v;
  program *p;

  t = this_task;

  /* clear any pending error condition now, then debug the task */
  t->error_pending = 0;

  if (!t->debug[(int) err])
    return;

  if (t->debug_func[(int) err] &&
      (func_exists(global_funcs,
		   t->debug_func[(int) err]) == FUNC_STALKCODE) &&
      !task_recursive(t, t->debug_func[(int) err])) {
    /* build the arguments for the function, and exec the function. */

    v = newlist(t->level+1);
    v->list[0] = make_error(NULL, err);
    x = t->level;
    while (t) {
      v->list[x - t->level+1] = debug_func(t);
      t = t->parent;
    }

    /* these two statements must use 'this_task' */
    p = find_program(global_funcs, this_task->debug_func[(int) err]);
    exec_task(this_task, p, v, 1);

    return;
  } else {
    t->type = TA_DYING;
    sprintf(debug_line, "program '%s' got error %s (%s), line %d",
	    t->program->name, err2token(err), err2string(err),
	    t->line_info->list[0]->num);
    (displaystr)(t->buf, debug_line);

    debug_dump_func(t);
    t = t->parent;

    while (t) {
      (displaystr)(t->buf, "");
      sprintf(debug_line, "called from program '%s', line %d",
	      t->program->name, t->line_info->list[0]->num);
      (displaystr)(t->buf, debug_line);

      debug_dump_func(t);
      t = t->parent;
    }

    return;
  }
}
