/*
 * stalk/functions.c
 * Copyright (c) 1993 Gerald Weiler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "include.h"
#include "functions.h"
#include "bi_func.h"
#include "stream.h"
#include "task.h"
#include "var.h"
#include "interpret.h"
#include "stuff.h"
#include "stalk.h"
#include "termt.h"
#include "termd.h"
#include "buffer.h"
#include "parser.h"

static prog_env *sub_set_prog P((prog_env *penv, char *name, program *p,
				 int builtin, bi_func func));

binding *global_bindings[256]; /* a global binding for each character */

prog_env *global_funcs;

/* global screen function pointers */
void (*setup_screen)   P((void));
void (*end_screen)     P((void));
void (*printchar)      P((buffer *buf, char c));
void (*displaystr)     P((buffer *buf, char *s));
void (*displaybold)    P((buffer *buf, char *s));
void (*handle_return)  P((buffer *buf));
void (*print_title)    P((buffer *buf));
void (*backspace)      P((buffer *buf));
void (*backward)       P((buffer *buf));
void (*forward)        P((buffer *buf));
void (*eol)            P((buffer *buf));
void (*bol)            P((buffer *buf));
void (*previous_line)  P((buffer *buf));
void (*next_line)      P((buffer *buf));
void (*kill_line)      P((buffer *buf));
void (*ctrlD)          P((buffer *buf));
void (*save_kb)        P((buffer *buf, keyboard_state *save));
void (*restore_kb)     P((buffer *buf, keyboard_state *save));
void (*redraw)         P((buffer *buf));
void (*page_down)      P((buffer *buf));
void (*page_up)        P((buffer *buf));
void (*scroll_down)    P((buffer *buf));
void (*scroll_up)      P((buffer *buf));
void (*save_buffer)    P((buffer *buf));
void (*connect_new)    P((buffer *buf));
int (*new_buffer)      P((buffer *prev, char *name, int Lines));

void setup_global_functions()
{
  binding **b;
  int x;
  char s[2];

  /* setup global functions */
  global_funcs = NULL;

  set_prog("readdir",        NULL, 1, bi_readdir);
  set_prog("stat",           NULL, 1, bi_stat);
  set_prog("getenv",         NULL, 1, bi_getenv);

  set_prog("rindex",         NULL, 1, bi_rindex);
  set_prog("index",          NULL, 1, bi_index);
  set_prog("crypt",          NULL, 1, bi_crypt);
  set_prog("random",         NULL, 1, bi_random);
  set_prog("strsub",         NULL, 1, bi_strsub);
  set_prog("length",         NULL, 1, bi_length);
  set_prog("ctime",          NULL, 1, bi_ctime);
  set_prog("time",           NULL, 1, bi_time);
  set_prog("tostr",          NULL, 1, bi_tostr);
  set_prog("tonum",          NULL, 1, bi_tonum);
  set_prog("typeof",         NULL, 1, bi_typeof);
  set_prog("vars",           NULL, 1, bi_vars);
  set_prog("char",           NULL, 1, bi_char);
  set_prog("quit",           NULL, 1, bi_quit);
  set_prog("version",        NULL, 1, bi_version);

  set_prog("new_function",   NULL, 1, bi_new_function);
  set_prog("rm_function",    NULL, 1, bi_rm_function);
  set_prog("list_function",  NULL, 1, bi_list_function);
  set_prog("functions",      NULL, 1, bi_functions);
  set_prog("eval",           NULL, 1, bi_eval);

  set_prog("get_input",      NULL, 1, bi_get_input);
  set_prog("set_input",      NULL, 1, bi_set_input);

  set_prog("ticks_left",     NULL, 1, bi_ticks_left);
  set_prog("set_ticks",      NULL, 1, bi_set_ticks);
  set_prog("default_ticks",  NULL, 1, bi_default_ticks);

  set_prog("die",            NULL, 1, bi_die);
  set_prog("suspend",        NULL, 1, bi_suspend);
  set_prog("task_id",        NULL, 1, bi_task_id);
  set_prog("kill_task",      NULL, 1, bi_kill_task);
  set_prog("queued_tasks",   NULL, 1, bi_queued_tasks);

  set_prog("bind_key",       NULL, 1, bi_bind_key);
  set_prog("unbind",         NULL, 1, bi_unbind);
  set_prog("bindings",       NULL, 1, bi_bindings);

  set_prog("popen",          NULL, 1, bi_popen);
  set_prog("fopen",          NULL, 1, bi_fopen);
  set_prog("copen",          NULL, 1, bi_copen);
  set_prog("sopen",          NULL, 1, bi_sopen);
  set_prog("accept",         NULL, 1, bi_accept);
  set_prog("close",          NULL, 1, bi_close);
  set_prog("read",           NULL, 1, bi_read);
  set_prog("write",          NULL, 1, bi_write);
  set_prog("set_timeout",    NULL, 1, bi_set_timeout);
  set_prog("capture_output", NULL, 1, bi_capture_output);
  set_prog("export",         NULL, 1, bi_export);
  set_prog("on_input",       NULL, 1, bi_on_input);
  set_prog("on_close",       NULL, 1, bi_on_close);
  set_prog("stream_ready",   NULL, 1, bi_stream_ready);
  set_prog("broadcast",      NULL, 1, bi_broadcast);
  set_prog("set_mode",       NULL, 1, bi_set_mode);
  set_prog("streams",        NULL, 1, bi_streams);

  set_prog("split_buffer",   NULL, 1, bi_split_buffer);
  set_prog("printbuffer",    NULL, 1, bi_printbuffer);
  set_prog("send_event",     NULL, 1, bi_send_event);
  set_prog("title",          NULL, 1, bi_title);
  set_prog("buffer_id",      NULL, 1, bi_buffer_id);

  switch (screen_mode) {
  case TERM_DUMB:
    d_setup_funcs();
    break;
  case TERM_TERMCAP:
    t_setup_funcs();
    break;
#ifdef HAVE_X11
  case TERM_X:
    x_setup_funcs();
    break;
#endif
  default:
    /* cannot use die() here */
    printf("huh? unknown terminal type %d\n", screen_mode);
    exit(-1);
    break;
  }

  b = global_bindings;
  for (x=0; x<256; x++)
    b[x] = NULL;

  /* this assumes ascii. sigh. */
  s[1] = '\0';
  for (*s=32; *s < 127; (*s)++)
    bind_key(NULL, s, binding_print, NULL, NULL, 1);

  bind_key(NULL, "\001", binding_builtin, bol, NULL, 1);          /* ctrl A */
  bind_key(NULL, "\002", binding_builtin, backward, NULL, 1);     /* ctrl B */
  bind_key(NULL, "\003", binding_builtin, w_quit, NULL, 1);       /* ctrl C */
  bind_key(NULL, "\004", binding_builtin, ctrlD, NULL, 1);        /* ctrl D */
  bind_key(NULL, "\005", binding_builtin, eol, NULL, 1);          /* ctrl E */
  bind_key(NULL, "\006", binding_builtin, forward, NULL, 1);      /* ctrl F */
  bind_key(NULL, "\007", binding_builtin, beep, NULL, 1);         /* ctrl G */
  bind_key(NULL, "\010", binding_builtin, backspace, NULL, 1);    /* ctrl H */
  bind_key(NULL, "\011", binding_builtin, do_nothing, NULL, 1);   /* ctrl I */
  bind_key(NULL, "\012", binding_builtin, handle_return, NULL, 1);/* ctrl J */
  bind_key(NULL, "\013", binding_builtin, kill_line, NULL, 1);    /* ctrl K */
  bind_key(NULL, "\014", binding_builtin, refresh, NULL, 1);      /* ctrl L */
  bind_key(NULL, "\015", binding_builtin, handle_return, NULL, 1);/* ctrl M */
  bind_key(NULL, "\016", binding_builtin, next_line, NULL, 1);    /* ctrl N */
  bind_key(NULL, "\017", binding_builtin, do_nothing, NULL, 1);   /* ctrl O */
  bind_key(NULL, "\020", binding_builtin, previous_line, NULL, 1);/* ctrl P */
  bind_key(NULL, "\021", binding_builtin, do_nothing, NULL, 1);   /* ctrl Q */
  bind_key(NULL, "\022", binding_builtin, page_down, NULL, 1);    /* ctrl R */
  bind_key(NULL, "\023", binding_builtin, do_nothing, NULL, 1);   /* ctrl S */
  bind_key(NULL, "\024", binding_builtin, page_up, NULL, 1);      /* ctrl T */
  bind_key(NULL, "\025", binding_builtin, scroll_up, NULL, 1);    /* ctrl U */
  bind_key(NULL, "\026", binding_builtin, do_nothing, NULL, 1);   /* ctrl V */
  bind_key(NULL, "\027", binding_builtin, do_nothing, NULL, 1);   /* ctrl W */
  bind_key(NULL, "\031", binding_builtin, scroll_down, NULL, 1);  /* ctrl Y */
  bind_key(NULL, "\032", binding_builtin, w_interrupt, NULL, 1);  /* ctrl Z */
  bind_key(NULL, "\177", binding_builtin, backspace, NULL, 1);    /* delete */
  bind_key(NULL, "\033N", binding_builtin, next_line, NULL, 1);    /* esc N */
  bind_key(NULL, "\033P", binding_builtin, previous_line, NULL, 1);/* esc P */
  bind_key(NULL, "\033n", binding_builtin, next_line, NULL, 1);    /* esc N */
  bind_key(NULL, "\033p", binding_builtin, previous_line, NULL, 1);/* esc P */
  bind_key(NULL, "\030\027", binding_builtin, save_buffer,
		     NULL, 1);                             /* ctrl X ctrl W */
  bind_key(NULL, "\030\022", binding_builtin, connect_new,
		     NULL, 1);                             /* ctrl X ctrl R */
  bind_key(NULL, "\0302", binding_builtin, split_buffer,
		     NULL, 1);                                  /* crtl X 2 */
  bind_key(NULL, "\030o", binding_builtin, next_buffer,
		     NULL, 1);                                  /* ctrl X o */
  bind_key(NULL, "\033[D", binding_builtin, backward,
		     NULL, 1);                               /* keypad left */
  bind_key(NULL, "\033[C", binding_builtin, forward,
		     NULL, 1);                              /* keypad right */
  bind_key(NULL, "\033[A", binding_builtin, previous_line,
		     NULL, 1);                                 /* keypad up */
  bind_key(NULL, "\033[B", binding_builtin, next_line,
		     NULL, 1);                               /* keypad down */
}

void bind_key(buf, key, type, func, func_name, global)
     buffer *buf;
     char *key;
     binding_type type;
     key_func func;
     char *func_name;
     int global;
{
  char *s;
  binding **b;
  int x, c;

  if ((type != binding_builtin) &&
      (type != binding_print) &&
      (type != binding_stalkcode))
    return;

  if ((type == binding_print) && (strlen(key) != 1))
    return;

  if (global)
    b = global_bindings;
  else {
    if (!buf->local_bindings) {
      buf->local_bindings = (binding **)malloc(256 * sizeof(binding *));
      for (x=0; x<256; x++)
	buf->local_bindings[x] = NULL;
    }

    b = buf->local_bindings;
  }

  s = key;

  while (*s) {
    c = *s;
    if (s[1]) {
      if (!b[c]) {
	b[c] = NEW(binding);
	b[c]->type = binding_none;
      } else if (b[c]->type == binding_stalkcode) {
	free(b[c]->name);
      }

      if (b[c]->type != binding_prefix) {
	b[c]->type = binding_prefix;
	b[c]->prefix = (binding **)malloc(256 * sizeof(binding *));
	for (x=0; x<256; x++)
	  b[c]->prefix[x] = NULL;
      }

      b = b[c]->prefix;
    } else {
      if (!b[c]) {
	b[c] = NEW(binding);
      } else if (b[c]->type == binding_prefix) {
	for (x=0; x<256; x++)
	  free_binding(b[c]->prefix[x]);
	free(b[c]->prefix);
      } else if (b[c]->type == binding_stalkcode) {
	free(b[c]->name);
      }

      b[c]->type = type;
      if (type == binding_builtin)
	b[c]->function = func;
      else if (type == binding_stalkcode)
	b[c]->name = mystrdup(func_name);
    }

    s++;
  }
}

void free_binding(b)
     binding *b;
{
  int x;

  if (!b)
    return;

  if (b->type == binding_stalkcode)
    free(b->name);

  if (b->type == binding_prefix) {
    for (x=0; x<256; x++)
      free_binding(b->prefix[x]);
    free(b->prefix);
  }

  free(b);
}

/*
 * returns 1 if the function is found and executed (the keystring is removed
 * from the string)
 * returns 0 if the string is a prefix and is not complete.
 * returns -1 if the string is not valid
 */
int check_bindings(b, buf, str)
     binding **b;
     buffer *buf;
     char *str;
{
  int x;
  char *s;
  var *v;

  x = 0;

  while (str[x]) {
    if (!b || !b[(int) str[x]])
      return -1;

    switch(b[(int) str[x]]->type) {
    case binding_print:
      (printchar)(buf, str[x]);
      s = str+x+1;
      while ((*str++ = *s++));
      return 1;
      break;
    case binding_builtin:
      (b[(int) str[x]]->function)(buf);
      s = str+x+1;
      while ((*str++ = *s++));
      return 1;
      break;
    case binding_stalkcode:
      if (!func_exists(global_funcs, b[(int) str[x]]->name))
	beep(buf);
      else {
	push_task(buf, find_program(global_funcs, b[(int) str[x]]->name),
		  v = newlist(0));
	var_free(v);
      }
      s = str+x+1;
      while ((*str++ = *s++));
      return 1;
      break;
    case binding_prefix:
      if (!str[x+1])
	return 0;
      b = b[(int) str[x]]->prefix;
      x++;
      break;
    default:
      return -1;
      break;
    }
  }
  return 0;
}

void check_functions(buf, str)
     buffer *buf;
     char *str;
{
  char *s;

  while (*str) {
    if (check_bindings(buf->local_bindings, buf, str) == 1)
      continue;

    switch (check_bindings(global_bindings, buf, str)) {
    case 1:
      continue;
      break;
    case -1:
      beep(buf);
      s = str+1;
      while ((*str++ = *s++));
      return;
      break;
    default:
      return;
      break;
    }
  }
}

var *function_list(t)
     task *t;
{
  var *r;
  int b, n;

  r = newlist(2);
  r->list[0] = newlist(count_funcs(global_funcs, 1));
  r->list[1] = newlist(count_funcs(global_funcs, 0));
  b = n = 0;
  function_list_sub(global_funcs, r, &b, &n);
  return r;
}

int count_funcs(p, builtin)
     prog_env *p;
     int builtin;
{
  if (!p)
    return 0;

  return count_funcs(p->left, builtin) + count_funcs(p->right, builtin) +
    (((!builtin) == (!p->builtin)) ? 1 : 0);
}

void function_list_sub(p, v, b, n)
     prog_env *p;
     var *v;
     int *b, *n;
{
  if (!p)
    return;
  function_list_sub(p->left, v, b, n);
  if (p->builtin)
    v->list[0]->list[(*b)++] = newstring(mystrdup(p->name));
  else
    v->list[1]->list[(*n)++] = newstring(mystrdup(p->name));
  function_list_sub(p->right, v, b, n);
}

static prog_env *sub_set_prog(penv, name, p, builtin, func)
     prog_env *penv;
     char *name;
     program *p;
     int builtin;
     bi_func func;
{
  int c;

  if (penv == NULL) {
    penv = NEW(prog_env);
    penv->left = penv->right = NULL;
    penv->name = mystrdup(name);
    penv->program = p;
    penv->builtin = builtin;
    penv->func = func;
    return penv;
  }

  c = strcmp(penv->name, name);
  if (!c) {
    if (penv->program->ref) {
      free(penv->program->name);
      penv->program->name = mystrdup("* deleted program *");
      penv->program->evaled = 1;
    } else
      free_program(penv->program);

    penv->program = p;
    penv->builtin = builtin;
    penv->func = func;
  } else if (c < 0)
    penv->right = sub_set_prog(penv->right, name, p, builtin, func);
  else
    penv->left = sub_set_prog(penv->left, name, p, builtin, func);

  return penv;
}

void set_prog(name, root, builtin, func)
     char *name;
     statement *root;
     int builtin;
     bi_func func;
{
  program *p;

  if (!builtin) {
    p = NEW(program);
    p->name = mystrdup(name);
    p->ref = 0;
    p->debug = 1;
    p->base = root;
    p->evaled = 0;
  } else
    p = NULL;

  global_funcs = sub_set_prog(global_funcs, name, p, builtin, func);
}

program *find_program(penv, name)
     prog_env *penv;
     char *name;
{
  int c;

  if (!penv)
    return NULL;

  if (!(c = strcmp(penv->name, name)) && penv->builtin)
    return NULL;
  else if (!c)
    return penv->program;
  else if (c < 0)
    return find_program(penv->right, name);
  else
    return find_program(penv->left, name);
}

/*
 * func_exists returns 0 if a function does not exist,
 * FUNC_BUILTIN if it is builtin, and FUNC_STALKCODE if it is not
 */
int func_exists(penv, name)
     prog_env *penv;
     char *name;
{
  int c;
  
  if (!penv)
    return 0;

  if (!(c = strcmp(penv->name, name)))
    return penv->builtin ? FUNC_BUILTIN : FUNC_STALKCODE;
  else if (c < 0)
    return func_exists(penv->right, name);
  else
    return func_exists(penv->left, name);
}

/*
 * builtin functions from the PL that act on function and keybinding
 * definitions
 */

void bi_functions(t, args)
     task *t;
     var *args;
{
  if (!args || (args->type != V_LIST) || args->num)
    in_push(t, make_error(t, E_ARGS));
  else
    in_push(t, function_list(t));
}

void bi_list_function(t, args)
     task *t;
     var *args;
{
  program *p;
  static var_type types[] = { V_STRING };
  Error e;

  if ((e = inv_args(args, 1, 1, types)) != E_NONE) {
    in_push(t, make_error(t, e));
    return;
  }

  p = find_program(global_funcs, args->list[0]->string);

  if (p == NULL)
    in_push(t, make_error(t, E_VERBNF));
  else
    in_push(t, unparse(p->base));
}

void bi_rm_function(t, args)
     task *t;
     var *args;
{
  in_push(t, NULL);
}

void bi_bind_key(t, args)
     task *t;
     var *args;
{
  key_func kf;
  int x;
  binding_type type;
  static var_type types[] = { V_STRING, V_STRING, V_NUM };
  Error e;
  int global;

  if ((e = inv_args(args, 2, 3, types)) != E_NONE) {
    in_push(t, make_error(t, e));
    return;
  }

  global = (args->num == 3) ? args->list[2]->num : 1;

  type = binding_stalkcode;
  kf = NULL;

  if (!mystrcasecmp("printchar", args->list[1]->string))
    type = binding_print;
  else
    for (x=0; t->buf->function_bindings[x].function_name != NULL; x++)
      if (!mystrcasecmp(t->buf->function_bindings[x].function_name,
			args->list[1]->string)) {
	type = binding_builtin;
	kf = t->buf->function_bindings[x].function;
      }

  bind_key(t->buf, args->list[0]->string, type, kf,
	   args->list[1]->string, global);

  in_push(t, NULL);
}

void bi_new_function(t, args)
     task *t;
     var *args;
{
  char *name;
  statement *root;
  static var_type types[] = { V_STRING, V_STRING };
  Error e;

  if ((e = inv_args(args, 2, 2, types)) != E_NONE) {
    in_push(t, make_error(t, e));
    return;
  }

  root = my_parse(t->buf, args->list[1]->string);
  name = args->list[0]->string;
  if (root == NULL)
    in_push(t, newnum(-1));
  else {
    set_prog(name, root, 0, NULL);
    in_push(t, newnum(0));
  }
}

void bi_bindings(t, args)
     task *t;
     var *args;
{
  in_push(t, NULL);
}

void bi_unbind(t, args)
     task *t;
     var *args;
{
  in_push(t, NULL);
}
