/*
 * stalk/functions.h
 */

#ifndef FUNCTIONS_H
#define FUNCTIONS_H

/* return values for func_exist */
#define FUNC_BUILTIN   1
#define FUNC_STALKCODE 2

/*
 * global function pointers
 */
extern void (*setup_screen)   P((void));
extern void (*end_screen)     P((void));
extern void (*printchar)      P((buffer *buf, char c));
extern void (*displaystr)     P((buffer *buf, char *s));
extern void (*displaybold)    P((buffer *buf, char *s));
extern void (*handle_return)  P((buffer *buf));
extern void (*print_title)    P((buffer *buf));
extern void (*backspace)      P((buffer *buf));
extern void (*backward)       P((buffer *buf));
extern void (*forward)        P((buffer *buf));
extern void (*eol)            P((buffer *buf));
extern void (*bol)            P((buffer *buf));
extern void (*previous_line)  P((buffer *buf));
extern void (*next_line)      P((buffer *buf));
extern void (*kill_line)      P((buffer *buf));
extern void (*ctrlD)          P((buffer *buf));
extern void (*save_kb)        P((buffer *buf, keyboard_state *save));
extern void (*restore_kb)     P((buffer *buf, keyboard_state *save));
extern void (*redraw)         P((buffer *buf));
extern void (*page_down)      P((buffer *buf));
extern void (*page_up)        P((buffer *buf));
extern void (*scroll_down)    P((buffer *buf));
extern void (*scroll_up)      P((buffer *buf));
extern void (*save_buffer)    P((buffer *buf));
extern void (*connect_new)    P((buffer *buf));
extern int (*new_buffer)      P((buffer *prev, char *name, int Lines));

extern prog_env *global_funcs;

void setup_global_functions   P((void));
void bind_key                 P((buffer *buf, char *key, binding_type type,
				 key_func func, char *func_name, int global));
void set_prog                 P((char *name, statement *root, int builtin,
				 bi_func func));
void free_binding             P((binding *b));
void check_functions          P((buffer *buf, char *str));
void function_list_sub        P((prog_env *p, var *v, int *b, int *n));

program *find_program         P((prog_env *penv, char *name));

int func_exists               P((prog_env *penv, char *name));
int check_bindings            P((binding **b, buffer *buf, char *str));
int count_funcs               P((prog_env *p, int builtin));

var *function_list            P((task *t));

void bi_new_function          P((task *t, var *args));
void bi_rm_function           P((task *t, var *args));
void bi_list_function         P((task *t, var *args));
void bi_functions             P((task *t, var *args));
void bi_bind_key              P((task *t, var *args));
void bi_unbind                P((task *t, var *args));
void bi_bindings              P((task *t, var *args));

#endif /* FUNCTIONS_H */
