/*
 * stalk/lines.c
 * Copyright (c) 1993 Gerald Weiler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "include.h"
#include "lines.h"
#include "stalk.h"
#include "stuff.h"

void add_line(lq, str, is_eol)
     line_queue *lq;
     char *str;
     int is_eol;
{
  line *lp;

  lp = lq->cursor;
  while(lp->next) {
    lp = lp->next;
  }

  lp->next = (line *)malloc(sizeof(line));
  lq->size++;

  lp->next->string = mystrdup(str);
  lp->next->prev = lp;
  lp->next->next = NULL;
  lp->next->eol = is_eol;
}

void advance_lq(lq)
     line_queue *lq;
{
  if (!lq->cursor->next)
    return;

  lq->cursor = lq->cursor->next;
  lq->pos++;
}

void retreat_lq(lq)
     line_queue *lq;
{
  if (!lq->cursor->prev)
    return;

  lq->cursor = lq->cursor->prev;
  lq->pos--;
}

void put_line_in_history(buf, str)
     buffer *buf;
     char *str;
{
  if (strlen(str) == 0)
    return;

  add_line(buf->history, str, 1);
}

void put_line_in_linequeue(buf, str, advance)
     buffer *buf;
     char *str;
     int advance;
{
  int x, y, size = strlen(str);
  static char *tmp = 0;

  if (tmp == 0)
    tmp = (char *)malloc(cols+5);

  if (size == 0) {
    add_line(buf->lines, "", 1);
    if (advance)
      advance_lq(buf->lines);
  } else {
    for(x=0; x<size; x+=cols) {
      /* # of characters to copy */
      y = (x+cols)<size ? cols : size-x;
      strncpy(tmp, str+x, y);
      tmp[y] = '\0';
      add_line(buf->lines, tmp, y<cols);
      if (advance)
	advance_lq(buf->lines);
    }
  }
}

line_queue *init_line_queue()
{
  line_queue *lq;

  lq = (line_queue *)malloc(sizeof(line_queue));
  lq->pos = lq->size = 0;
  lq->cursor = lq->lines = (line *)malloc(sizeof(line));
  lq->lines->string = NULL;
  lq->lines->next = lq->lines->prev = NULL;
  return lq;
}
