%{
/*
 * stalk/parser.y
 * Copyright (c) 1993 Gerald Weiler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "include.h"
#include "parser.h"
#include "var.h"
#include "stuff.h"
#include "functions.h"

#ifdef NEED_ALLOCA
#include <alloca.h>
#endif

/* contains the program/statement to be parsed */
static char parse_buffer[9999];

/* length and current position in parse_buffer */
static int blen, bpos;

/* used as temporary storage */
static char lbuff[1000];
static int llen;

/* the buffer to print to */
static buffer *cbuf;

static char *ctemp;

static int lineno;
static char err;
static statement *prgm;
static var *vtemp;
static statement *stemp, *stemp2;
static exprstruct *etemp;

static string_list *slist;

#if 0
static char precedence[50];
#endif

static void error P((char *s1, char *s2));
%}

%union {
  int integer;
  char *string;
  statement *stmt;
  exprstruct *expr;
  Error err;
}

%token <integer> T_INTEGER
%token <err> T_ERROR
%token <string> T_STRING T_ID
%token T_GE T_GT T_LE T_LT T_EQ T_NE T_NOT T_OR T_AND T_TO T_ASSIGN
%token T_IF T_ELSEIF T_ELSE T_ENDIF T_FOR T_IN T_ENDFOR T_RETURN
%token T_WHILE T_ENDWHILE T_FORK T_ENDFORK T_EXIT
%token T_ONERROR T_CALL T_IGNORE T_FATAL

%type <stmt> statement statements elseifs elsepart
%type <expr> expr arglist nearglist ne_errlist err_action

%right ','
%right T_ASSIGN
%nonassoc '?' '|'
%left T_AND T_OR
%left T_GE T_GT T_LE T_LT T_EQ T_NE T_IN
%left '+' '-'
%left '*' '/' '%'
%left T_UMINUS T_NOT
%nonassoc '$' '['

%%

stalkcode: statements
               { prgm = $1; }

statements:   /* nothing */
               { $$ = NULL; }
       | statements statement
               { $$ = append_statements($1, $2); }
       ;

statement: T_IF '(' expr ')' statements elseifs elsepart T_ENDIF
               { $$ = append_statements(make_line_stmt($3->lvars, $3->gvars,
						       strlen($3->upexpr)+6,
						       "if (%s)",
						       $3->upexpr),
					$3->s);
		 free($3->upexpr);
		 free($3);
		 $$ = append_statements($$, stemp = make_statement(MN_IF));
		 stemp->next = $5;
		 append_statements(stemp, make_statement(MN_JUMP));
		 append_statements(stemp, $6);
		 append_statements(stemp, $7);
		 stemp2 = append_statements(make_line_stmt(NULL, NULL, 6,
							   "endif"),
					    make_statement(MN_ENDIF));
		 append_statements(stemp, stemp2);

		 /* chain everything together */
		 if ($6)
		   stemp->jump = $6;
		 else if ($7)
		   stemp->jump = $7;
		 else
		   stemp->jump = stemp2;

		 if ($6) {
		   stemp = $6;
		   while (stemp->type != MN_ELSEIF || stemp->jump)
		     stemp = stemp->next;
		   if ($7)
		     stemp->jump = $7;
		   else
		     stemp->jump = stemp2;
		 }

		 stemp = $$;
		 while (stemp) {
		   if (stemp->type == MN_JUMP && !stemp->jump)
		     stemp->jump = stemp2;
		   stemp = stemp->next;
		 }
	       }
       | T_FOR T_ID T_IN '(' expr ')' statements T_ENDFOR
               { string_list *sl;

		 sl = NEW(string_list);
		 sl->string = mystrdup($2);
		 sl->next = $5->lvars;
		 $$ = make_line_stmt(sl, $5->gvars,
				     strlen($2)+strlen($5->upexpr)+11,
				     "for %s in (%s)", $2, $5->upexpr);
		 $$ = append_statements($$, make_statement(MN_ID));
		 $$->next->value = newlabel($2);
		 $$->next->next = make_statement(MN_VALUE);
		 $$->next->next->value = newnum(0);
		 $$->next->next->next = $5->s;
		 free($5->upexpr);
		 free($5);
		 append_statements($$, stemp = make_statement(MN_FOR_LIST));
		 stemp->next = $7;
		 stemp2 = append_statements(make_line_stmt(NULL, NULL, 7,
							   "endfor"),
					    make_statement(MN_ENDFOR));
		 append_statements(stemp, stemp2);

		 stemp->jump = stemp2->next;
		 stemp2->next->jump = stemp;

		 /* search for exits */
		 for (stemp = $$; stemp; stemp = stemp->next)
		   if ((stemp->type == MN_EXIT) && !stemp->jump &&
		       !strcmp(stemp->value->string, $2))
		     stemp->jump = stemp2;
	       }
       | T_FOR T_ID T_IN '[' expr T_TO expr ']' statements T_ENDFOR
	       { string_list *sl1, *sl2;

		 sl1 = NEW(string_list);
		 sl1->string = mystrdup($2);
		 sl1->next = $5->lvars;
		 while (sl1->next)
		   sl1 = sl1->next;
		 sl1->next = $7->lvars;
		 sl2 = $5->gvars;
		 if (sl2) {
		   while (sl2->next)
		     sl2 = sl2->next;
		   sl2->next = $7->gvars;
		 } else
		   sl2 = $7->gvars;

		 $$ = make_line_stmt(sl1, sl2, strlen($2)+strlen($5->upexpr)+
				     strlen($7->upexpr)+13,
				     "for %s in [%s..%s]", $2, $5->upexpr,
				     $7->upexpr),
		 $$ = append_statements($$, make_statement(MN_ID));
		 $$->next->value = newlabel($2);
		 $$->next->next = $5->s;
		 free($5->upexpr);
		 free($5);
		 $$ = append_statements($$, $7->s);
		 free($7->upexpr);
		 free($7);
		 $$ = append_statements($$,
					stemp = make_statement(MN_FOR_RANGE));
		 stemp->next = $9;
		 stemp2 = append_statements(make_line_stmt(NULL, NULL, 7,
							   "endfor"),
					    make_statement(MN_ENDFOR));
		 stemp = append_statements(stemp, stemp2);
		 stemp->jump = stemp2->next;
		 stemp2->next->jump = stemp;

		 /* search for exits */
		 for (stemp = $$; stemp; stemp = stemp->next)
		   if ((stemp->type == MN_EXIT) && !stemp->jump &&
		       !strcmp(stemp->value->string, $2))
		     stemp->jump = stemp2;
	       }
       | T_WHILE '(' expr ')' statements T_ENDWHILE
               { $$ = append_statements(make_line_stmt($3->lvars, $3->gvars,
						       strlen($3->upexpr) + 9,
						       "while (%s)",
						       $3->upexpr),
					$3->s);
                 $$ = append_statements($$,
					stemp = make_statement(MN_WHILE));
                 free($3->upexpr);
                 free($3);
		 stemp = append_statements(stemp, $5);
		 stemp2 = append_statements(make_line_stmt(NULL, NULL, 9,
							   "endwhile"),
					    make_statement(MN_ENDWHILE));
		 stemp = append_statements(stemp, stemp2);
		 stemp->jump = stemp2->next;
		 stemp2->next->jump = $$;
	       }
       | T_WHILE T_ID '(' expr ')' statements T_ENDWHILE
               { $$ = make_line_stmt($4->lvars, $4->gvars,
				     strlen($2) + strlen($4->upexpr) + 10,
				     "while %s (%s)", $2, $4->upexpr);
		 $$ = append_statements($$, $4->s);
                 $$ = append_statements($$, stemp = make_statement(MN_WHILE));
		 stemp->value = newlabel($2);
                 free($4->upexpr);
                 free($4);
		 stemp = append_statements(stemp, $6);
		 stemp2 = append_statements(make_line_stmt(NULL, NULL, 9,
							   "endwhile"),
					    make_statement(MN_ENDWHILE));
		 stemp = append_statements(stemp, stemp2);
		 stemp->jump = stemp2->next;
		 stemp2->next->jump = $$;

		 /* search for exits */
		 for (stemp = $$; stemp; stemp = stemp->next)
		   if ((stemp->type == MN_EXIT) && !stemp->jump &&
		       !strcmp(stemp->value->string, $2))
		     stemp->jump = stemp2;
	       }
       | T_FORK '(' expr ')' statements T_ENDFORK
               { $$ = append_statements(make_line_stmt($3->lvars, $3->gvars,
						       strlen($3->upexpr)+8,
						       "fork (%s)",
						       $3->upexpr),
					$3->s);
		 $$ = append_statements($$, stemp = make_statement(MN_FORK));
		 free($3->upexpr);
		 free($3);
		 stemp = append_statements(stemp, $5);
		 stemp2 = append_statements(make_line_stmt(NULL, NULL, 8,
							   "endfork"),
					    make_statement(MN_ENDFORK));
		 stemp2 = append_statements(stemp2, make_statement(MN_NOOP));
		 stemp = append_statements(stemp, stemp2);
		 stemp->jump = stemp2->next->next;
	       }
       | T_ONERROR ne_errlist err_action ';'
               { $$ = make_line_stmt($3->lvars, $3->gvars,
				     strlen($2->upexpr)+strlen($3->upexpr)+12,
				     "on_error %s %s;",
				     $2->upexpr, $3->upexpr);
		 $$ = append_statements($$, make_statement(MN_DEBUG));
		 stemp = append_statements(make_statement(MN_END_DEBUG),
					   $3->s);
		 $$ = append_statements($$, append_statements($2->s, stemp));
		 free($2->upexpr);
		 free($2);
		 free($3->upexpr);
		 free($3);
	       }
       | T_RETURN ';'
               { $$ = append_statements(make_line_stmt(NULL, NULL, 8,
						       "return;"),
					make_statement(MN_EMPTY_RETURN));
	       }
       | T_RETURN expr ';'
               { $$ = append_statements(make_line_stmt($2->lvars, $2->gvars,
						       strlen($2->upexpr)+9,
						       "return %s;",
						       $2->upexpr),
					$2->s);
		 $$ = append_statements($$, make_statement(MN_RETURN));
		 free($2->upexpr);
		 free($2);
	       }
       | T_EXIT T_ID ';'
               { string_list *sl;

		 sl = NEW(string_list);
		 sl->string = mystrdup($2);
		 sl->next = NULL;
		 $$ = append_statements(make_line_stmt(sl, NULL, strlen($2)+7,
						       "exit %s;", $2),
					make_statement(MN_EXIT));
		 vtemp = newstring($2);
		 $$->next->value = vtemp;
	       }
       | expr ';'
               { $$ = append_statements(make_line_stmt($1->lvars, $1->gvars,
						       strlen($1->upexpr)+2,
						       "%s;",
						       $1->upexpr),
					$1->s);
		 $$ = append_statements($$, make_statement(MN_ENDEXPR));
		 free($1->upexpr);
		 free($1);
	       }
       | ';'
               { $$ = append_statements(make_line_stmt(NULL, NULL, 2, ";"),
					make_statement(MN_NOOP));
	       }
       ;

elseifs:       { $$ = NULL; }
       | elseifs T_ELSEIF '(' expr ')' statements
               { stemp = make_line_stmt($4->lvars, $4->gvars,
					strlen($4->upexpr)+10,
					"elseif (%s)", $4->upexpr);
		 stemp = append_statements(stemp, $4->s);
		 free($4->upexpr);
		 free($4);
		 stemp = append_statements(stemp,
					   stemp2 = make_statement(MN_ELSEIF));
		 stemp2->next = $6;
		 stemp2 = append_statements(stemp2, make_statement(MN_JUMP));
		 if ($1) {
		   stemp2 = $1;
		   while ((stemp2->type != MN_ELSEIF) || stemp2->jump)
		     stemp2 = stemp2->next;
		   stemp2->jump = stemp;
		 }
		 $$ = append_statements($1, stemp);
	       }
       ;

elsepart: 
  /* nothing */
               { $$ = NULL; }
       | T_ELSE statements
               { $$ = append_statements(make_line_stmt(NULL, NULL, 5, "else"),
					make_statement(MN_ELSE));
		 $$ = append_statements($$, $2);
	       }
       ;

expr:    T_INTEGER
	       { $$ = NEW(exprstruct);
		 $$->lhs = 0;
		 $$->lvars = NULL;
		 $$->gvars = NULL;
		 $$->s = make_statement(MN_VALUE);
		 vtemp = newnum($1);
		 $$->s->value = vtemp;
		 $$->upexpr = (char *)malloc(10);
		 sprintf($$->upexpr, "%d", $1);
	       }
       | T_ID
	       { $$ = NEW(exprstruct);
		 $$->lhs = 1;
		 $$->lvars = NEW(string_list);
		 $$->lvars->string = mystrdup($1);
		 $$->lvars->next = NULL;
		 $$->gvars = NULL;
		 $$->s = make_statement(MN_ID);
		 vtemp = newstring($1);
		 $$->s->value = vtemp;
		 $$->s = append_statements($$->s,
					   make_statement(MN_LOCAL_VAR));
		 $$->upexpr = mystrdup($1);
	       }
       | '$' T_ID
	       { $$ = NEW(exprstruct);
		 $$->lhs = 1;
		 $$->gvars = NEW(string_list);
		 $$->gvars->string = mystrdup($2);
		 $$->gvars->next = NULL;
		 $$->lvars = NULL;
		 $$->s = make_statement(MN_ID);
		 vtemp = newstring($2);
		 $$->s->value = vtemp;
		 $$->s = append_statements($$->s,
					   make_statement(MN_GLOBAL_VAR));
		 $$->upexpr = (char *)malloc(strlen($2)+2);
		 sprintf($$->upexpr, "$%s", $2);
	       }
       | T_STRING
	       { char *s1, *s2;
		 $$ = NEW(exprstruct);
		 $$->lhs = 0;
		 $$->lvars = NULL;
		 $$->gvars = NULL;
		 $$->s = make_statement(MN_VALUE);
		 vtemp = newstring($1);
		 $$->s->value = vtemp;

		 /* [de?]mangle the string */
		 s1 = $$->upexpr = (char *)malloc(3+2*strlen($1));
		 *s1++ = '\"';
		 s2 = $1;
		 while (*s2) {
		   if ((*s2 == '\\') || (*s2 == '\"'))
		     *s1++ = '\\';
		   *s1++ = *s2++;
		 }
		 *s1++ = '\"';
		 *s1 = '\0';
	       }
       | T_ERROR
	       { $$ = NEW(exprstruct);
		 $$->lhs = 0;
		 $$->lvars = NULL;
		 $$->gvars = NULL;
		 $$->s = make_statement(MN_VALUE);
		 $$->s->value = make_error(NULL, $1);
		 $$->upexpr = mystrdup(err2token($1));
	       }
       | expr '+' expr
               { $$ = make_expr(MN_ADD, $1, $3, NULL); }
       | expr '-' expr
               { $$ = make_expr(MN_SUB, $1, $3, NULL); }
       | expr '*' expr
               { $$ = make_expr(MN_MUL, $1, $3, NULL); }
       | expr '/' expr
               { $$ = make_expr(MN_DIV, $1, $3, NULL); }
       | expr '%' expr
               { $$ = make_expr(MN_MOD, $1, $3, NULL); }
       | expr T_GT expr
               { $$ = make_expr(MN_GT, $1, $3, NULL); }
       | expr T_GE expr
               { $$ = make_expr(MN_GE, $1, $3, NULL); }
       | expr T_LT expr
               { $$ = make_expr(MN_LT, $1, $3, NULL); }
       | expr T_LE expr
               { $$ = make_expr(MN_LE, $1, $3, NULL); }
       | expr T_EQ expr
               { $$ = make_expr(MN_EQ, $1, $3, NULL); }
       | expr T_NE expr
               { $$ = make_expr(MN_NE, $1, $3, NULL); }
       | expr T_IN expr
               { $$ = make_expr(MN_IN, $1, $3, NULL); }
       | expr T_OR expr
               { $$ = make_expr(MN_OR, $1, $3, NULL); }
       | expr T_AND expr
               { $$ = make_expr(MN_AND, $1, $3, NULL); }
       | T_NOT expr
               { $$ = make_expr(MN_NOT, $2, NULL, NULL); }
       | '-' expr %prec T_UMINUS
               { $$ = make_expr(MN_NEG, $2, NULL, NULL); }
       | '(' expr ')'
               { $2->lhs = 0;
		 $$ = $2; }
       | expr '[' expr ']'
               { $$ = make_expr(MN_ELEMENT, $1, $3, NULL); }
       | expr '[' expr T_TO expr ']'
               { $$ = make_expr(MN_ELEMENT_RANGE, $1, $3, $5); }
       | expr '[' T_TO expr ']'
               { exprstruct *e;
		 e = NEW(exprstruct);
		 e->lhs = 0;
		 e->s = make_statement(MN_VALUE);
		 e->s->value = NULL;
		 e->upexpr = mystrdup("");
		 e->lvars = NULL;
		 e->gvars = NULL;
		 $$ = make_expr(MN_ELEMENT_RANGE, $1, e, $4);
		 $$->lhs = 0;
	       }
       | expr '[' expr T_TO ']'
               { exprstruct *e;
		 e = NEW(exprstruct);
		 e->lhs = 0;
		 e->s = make_statement(MN_VALUE);
		 e->s->value = NULL;
		 e->upexpr = mystrdup("");
		 e->lvars = NULL;
		 e->gvars = NULL;
		 $$ = make_expr(MN_ELEMENT_RANGE, $1, $3, e);
		 $$->lhs = 0;
	       }
       | expr T_ASSIGN expr
               { if (!$1->lhs)
		   yyerror("Illegal lhs of assignment;");
		 if ($1->s->next->type == MN_LOCAL_VAR)
		   $1->s->next->type = MN_LOCAL_ASSIGN;
		 else
		   $1->s->next->type = MN_GLOBAL_ASSIGN;

		 ctemp = (char *)malloc(strlen($1->upexpr) +
					strlen($3->upexpr) + 4);
		 sprintf(ctemp, "(%s=%s)", $1->upexpr, $3->upexpr);
		 free($1->upexpr);
		 free($3->upexpr);
		 $$ = $1;
		 $$->upexpr = ctemp;
		 $$->lhs = 0;

		 if ($$->lvars) {
		   slist = $$->lvars;
		   while (slist->next)
		     slist = slist->next;
		   slist->next = $3->lvars;
		 } else
		   $$->lvars = $3->lvars;

		 if ($$->gvars) {
		   slist = $$->gvars;
		   while (slist->next)
		     slist = slist->next;
		   slist->next = $3->gvars;
		 } else
		   $$->gvars = $3->gvars;

		 $$->s = append_statements($$->s, $3->s);
		 free($3);
		 $$->s = append_statements($$->s,
					   make_statement(MN_END_ASSIGN));
	       }
       | '{' arglist '}'
               { $$->s = append_statements(($$ = $2)->s,
					   make_statement(MN_LIST));
		 ctemp = (char *)malloc(strlen($$->upexpr) + 3);
		 sprintf(ctemp, "{%s}", $$->upexpr);
		 free($$->upexpr);
		 $$->upexpr = ctemp;
	       }
       | T_ID '(' arglist ')'
               { etemp = NEW(exprstruct);
		 etemp->s = make_statement(MN_ID);
		 etemp->s->value = newstring($1);
		 etemp->upexpr = mystrdup($1);
		 etemp->lvars = NULL;
		 etemp->gvars = NULL;
		 $$ = make_expr(MN_FUNCTION, etemp, $3, NULL);
	       }
       | '(' expr ')' '(' arglist ')'
               { $$ = make_expr(MN_INDIRECT_FUNCTION, $2, $5, NULL); }
       | expr '?' expr '|' expr
               { $$ = make_expr(MN_CONDITION, $1, $3, $5); }
       ;

arglist:
	       { $$ = NEW(exprstruct);
		 $$->lhs = 0;
		 $$->lvars = NULL;
		 $$->gvars = NULL;
		 $$->upexpr = mystrdup("");
		 $$->s = make_statement(MN_LIST_EMPTY);
	       }
       | nearglist
               { $$ = $1; }
       ;

nearglist: expr
               { $$ = make_expr(MN_LIST_ELEMENT, $1, NULL, NULL); }
       | '@' expr
               { $$ = make_expr(MN_LIST_SPLICE, $2, NULL, NULL); }
       | nearglist ',' nearglist
               { $$ = make_expr(MN_LIST_COMBINE, $1, $3, NULL); }
       ;

ne_errlist: T_ERROR
	       { $$ = NEW(exprstruct);
		 $$->lhs = 0;
		 $$->lvars = NULL;
		 $$->gvars = NULL;
		 $$->s = make_statement(MN_VALUE);
		 $$->s->value = make_error(NULL, $1);
		 $$->upexpr = mystrdup(err2token($1));
	       }
       | ne_errlist ',' ne_errlist
	       { $$ = $1;
		 $$->s->next = $3->s;
		 ctemp = (char *)malloc(strlen($1->upexpr) +
					strlen($3->upexpr) + 3);
		 sprintf(ctemp, "%s, %s", $1->upexpr, $3->upexpr);
		 free($1->upexpr);
		 free($3->upexpr);
		 free($3);
		 $$->upexpr = ctemp;
	       }
       ;

err_action: T_CALL expr
	       { $$ = $2;
		 $$->lhs = 0;
		 $$->s = append_statements($$->s, make_statement(MN_CALL));
		 ctemp = (char *)malloc(strlen($2->upexpr) + 6);
		 sprintf(ctemp, "call %s", $2->upexpr);
		 free($2->upexpr);
		 $$->upexpr = ctemp;
	       }
       | T_IGNORE
	       { $$ = NEW(exprstruct);
		 $$->lhs = 0;
		 $$->lvars = NULL;
		 $$->gvars = NULL;
		 $$->s = make_statement(MN_IGNORE);
		 $$->upexpr = mystrdup("ignore");
	       }
       | T_FATAL
	       { $$ = NEW(exprstruct);
		 $$->lhs = 0;
		 $$->lvars = NULL;
		 $$->gvars = NULL;
		 $$->s = make_statement(MN_FATAL);
		 $$->upexpr = mystrdup("fatal");
	       }
       ;

%%

int yylex()
{
  static tokenentry tokentable[] = {
    {"if",         T_IF},
    {"elseif",     T_ELSEIF},
    {"else",       T_ELSE},
    {"endif",      T_ENDIF},
    {"for",        T_FOR},
    {"in",         T_IN},
    {"endfor",     T_ENDFOR},
    {"while",      T_WHILE},
    {"endwhile",   T_ENDWHILE},
    {"return",     T_RETURN},
    {"fork",       T_FORK},
    {"endfork",    T_ENDFORK},
    {"exit",       T_EXIT},
    {"on_error",   T_ONERROR},
    {"call",       T_CALL},
    {"ignore",     T_IGNORE},
    {"fatal",      T_FATAL},

    {"e_none",     T_ERROR,    E_NONE},
    {"e_varnf",    T_ERROR,    E_VARNF},
    {"e_type",     T_ERROR,    E_TYPE},
    {"e_div",      T_ERROR,    E_DIV},
    {"e_verbnf",   T_ERROR,    E_VERBNF},
    {"e_maxrec",   T_ERROR,    E_MAXREC},
    {"e_range",    T_ERROR,    E_RANGE},
    {"e_args",     T_ERROR,    E_ARGS},
    {"e_invarg",   T_ERROR,    E_INVARG},
    {"e_ticks",    T_ERROR,    E_TICKS},

    {"\0",         0}
  };
  int c, x, base, done, digit;

  do {
    if ((c=lexget()) == '\n')
      lineno++;
  } while (isspace(c));

  if (c==EOF) {
    return c;
  } if (isdigit(c)) { /* number */
    digit = 0;

    if (c != '0') {
      base = 10;
      yylval.integer = c-'0';
    } else {
      c = lexget();
      if (c == 'x' || c == 'X') {
	base = 16;
	yylval.integer = 0;
      } else if (isdigit(c) && (c < '8')) {
	base = 8;
	yylval.integer = c - '0';
      } else {
	if (c != EOF)
	  lexunget();
	yylval.integer = 0;
	return T_INTEGER;
      }
    }

    done = 0;
    while (1) {
      c = lexget();
      switch (base) {
      case 8:
	if ((c < '0') || (c > '7')) {
	  done = 1;
	  break;
	}
	digit = c - '0';
	break;
      case 10:
	if (!isdigit(c)) {
	  done = 1;
	  break;
	}
	digit = c - '0';
	break;
      case 16:
	if (!isxdigit(c)) {
	  done = 1;
	  break;
	}
	if (isdigit(c))
	  digit = c - '0';
	else
	  digit = tolower(c) - 'a' + 10;
      default:
	done = 1;
	break;
      }
      if (done)
	break;
      yylval.integer *= base;
      yylval.integer += digit;
    }

    if (c != EOF)
      lexunget();
    return T_INTEGER;
  } else if (isalpha(c) || c == '_') { /* identifier or reserved word */
    llen = 0;
    lbuff[llen++] = c;
    while (isalnum(c=lexget()) || c == '_')
      lbuff[llen++] = c;
    lexunget();
    lbuff[llen] = '\0';

    /* check for one of the reserved words */
    for (x=0; tokentable[x].string[0] != '\0'; x++) 
      if (!mystrcasecmp(lbuff, tokentable[x].string)) {
	if (tokentable[x].value != T_ERROR) {
	  return tokentable[x].value;
	}
	yylval.err = tokentable[x].err;
	return T_ERROR;
      }
    yylval.string = mystrdup(lbuff);
    return T_ID;
  } else if (c == '"') {
    llen = 0;
    while (1) {
      if ((c=lexget()) == '"')
	break;
      if (c=='\\')
	c=lexget();
      if (c=='\n' || c=='\0' || c==EOF) {
	yyerror("missing quote"); break;
      }
      lbuff[llen++] = c;
    }
    lbuff[llen] = '\0';
    yylval.string = mystrdup(lbuff);
    return T_STRING;
  }
  switch (c) {
  case '>':
    if (lexget() == '=') {
      return T_GE;
    } else {
      lexunget();
      return T_GT;
    }
  case '<':
    if (lexget() == '=') {
      return T_LE;
    } else {
      lexunget();
      return T_LT;
    }
  case '=':
    if (lexget() == '=') {
      return T_EQ;
    } else {
      lexunget();
      return T_ASSIGN;
    }
  case '!':
    if (lexget() == '=') {
      return T_NE;
    } else {
      lexunget();
      return T_NOT;
    }
  case '|':
    if (lexget() == '|') {
      return T_OR;
    } else {
      lexunget();
      return '|';
    }
  case '&':
    if (lexget() == '&') {
      return T_AND;
    } else {
      lexunget();
      return '&';
    }
  case '.':
    if (lexget() == '.') {
      return T_TO;
    } else {
      lexunget();
      return '.';
    }
  default:
    return c;
  }
}

int lexget()
{
  if (bpos<blen)
    return parse_buffer[bpos++];
  else {
    bpos++;
    return EOF;
  }
}

void lexunget()
{
  if (bpos)
    bpos--;
}

statement *my_parse(buf, s)
     buffer *buf;
     char *s;
{
  int x;

  cbuf=buf;
  bpos=0;
  lineno=1;
  err=0;
  strncpy(parse_buffer, s, (blen=strlen(s)));
  parse_buffer[blen] = '\0';
  yyparse();
  if (err) {
    (displaystr)(cbuf, "error in program.\n");
    /***/ /* free the program here */
    return NULL;
  }

  /* search for exits left over */
  for (stemp = prgm; stemp; stemp = stemp->next)
    if ((stemp->type == MN_EXIT) && !stemp->jump) {
      char warning[999];
      sprintf(warning, "EXIT labeled %s not within FOR or WHILE labeled %s.\n",
	      stemp->value->string, stemp->value->string);
      (displaystr)(cbuf, warning);
      free_statements(prgm);
      return NULL;
    }

  /* number the lines */
  for (stemp = prgm, x=0; stemp; stemp = stemp->next)
    if (stemp->type == MN_LINE)
      stemp->value->list[0]->num = ++x;

  return prgm;
}

int yyerror(s)
     char *s;
{
  error(s, NULL);
  return 0;
}

static void error(s1, s2)
     char *s1, *s2;
{
  char string[500];

  err++;
  if (s2)
    sprintf(string, "Line %d: %s%s\n", lineno, s1, s2);
  else
    sprintf(string, "Line %d: %s\n", lineno, s1);
  (displaystr)(cbuf, string);
}

void init_prec()
{
/*
  precedence[MN_ELEMENT] = 7;

  precedence[MN_NOT] = 6;
  precedence[MN_NEG] = 6;

  precedence[MN_MUL] = 5;
  precedence[MN_DIV] = 5;
  precedence[MN_MOD] = 5;

  precedence[MN_ADD] = 4;
  precedence[MN_SUB] = 4;

  precedence[MN_EQ] = 3;
  precedence[MN_NE] = 3;
  precedence[MN_GE] = 3;
  precedence[MN_GT] = 3;
  precedence[MN_LE] = 3;
  precedence[MN_LT] = 3;

  precedence[MN_OR] = 2;
  precedence[MN_AND] = 2;

  precedence[MN_ASSIGN] = 1;

  precedence[MN_NONE] = 0;
  precedence[MN_IN] = 0;
  precedence[MN_TO] = 0;
  precedence[MN_MAKE_LIST] = 0;
  precedence[MN_SINGLE_ELT] = 0;
  precedence[MN_COMBINE] = 0;
  precedence[MN_SPLICE] = 0;
*/
}

exprstruct *make_expr(op, e1, e2, e3)
     mnemonic op;
     exprstruct *e1, *e2, *e3;
{
  exprstruct *e;
  statement *s1, *s2;
  string_list *s_l;
  char *c1, *c2, *c3, *c;

  e = NEW(exprstruct);
  if (op == MN_ELEMENT || op == MN_ELEMENT_RANGE)
    e->lhs = e1->lhs;
  else
    e->lhs = 0;

  e->lvars = e1->lvars;
  e->gvars = e1->gvars;
  s1 = make_statement(op);
  s2 = NULL;

  e->s = e1->s;
  c = NULL;
  c1 = e1->upexpr;
  c2 = NULL;
  c3 = NULL;
  free(e1);

  if (op == MN_OR || op == MN_AND) {
    if (op == MN_OR)
      append_statements(e->s, s2 = make_statement(MN_OR1));
    else if (op == MN_AND)
      append_statements(e->s, s2 = make_statement(MN_AND1));
  } else if (op == MN_CONDITION) {
    append_statements(e->s, s1);
    s1->jump = e3->s;
  }

  if (e2) {
    /* chain e2's local vars to e1's */
    if (e->lvars && e2->lvars) {
      s_l = e->lvars;
      while (s_l->next)
	s_l = s_l->next;
      s_l->next = e2->lvars;
    } else if (e2->lvars)
      e->lvars = e2->lvars;

    /* chain e2's global vars to e1's */
    if (e->gvars && e2->gvars) {
      s_l = e->gvars;
      while (s_l->next)
	s_l = s_l->next;
      s_l->next = e2->gvars;
    } else if (e2->gvars)
      e->gvars = e2->gvars;

    append_statements(e->s, e2->s);
    c2 = e2->upexpr;
    free(e2);

    if (op == MN_CONDITION)
      append_statements(e->s, s2 = make_statement(MN_JUMP));
  }

  if (e3) {
    /* chain e3's local vars to e1's */
    if (e->lvars && e3->lvars) {
      s_l = e->lvars;
      while (s_l->next)
	s_l = s_l->next;
      s_l->next = e3->lvars;
    } else if (e3->lvars)
      e->lvars = e3->lvars;

    /* chain e3's global vars to e1's */
    if (e->gvars && e3->gvars) {
      s_l = e->gvars;
      while (s_l->next)
	s_l = s_l->next;
      s_l->next = e3->gvars;
    } else if (e3->gvars)
      e->gvars = e3->gvars;

    append_statements(e->s, e3->s);
    c3 = e3->upexpr;
    free(e3);

    if (op == MN_CONDITION)
      append_statements(e->s, s1 = make_statement(MN_NOOP));
  }

  if (op != MN_CONDITION)
    append_statements(e->s, s1);

  if ((op == MN_OR) || (op == MN_AND) || (op == MN_CONDITION))
    s2->jump = s1;

  /* handle unparsing the expression */
  switch (op) {
  case MN_ELEMENT:
    c = (char *)malloc(strlen(c1)+strlen(c2)+3);
    sprintf(c, "%s[%s]", c1, c2);
    break;
  case MN_ELEMENT_RANGE:
    c = (char *)malloc(strlen(c1)+strlen(c2)+strlen(c3)+5);
    sprintf(c, "%s[%s..%s]", c1, c2, c3);
    break;
  case MN_LIST_ELEMENT:
    c = c1;
    c1 = NULL;
    break;
  case MN_LIST_COMBINE:
    c = (char *)malloc(strlen(c1)+strlen(c2)+3);
    sprintf(c, "%s, %s", c1, c2);
    break;
  case MN_LIST_SPLICE:
    c = (char *)malloc(strlen(c1)+2);
    sprintf(c, "@%s", c1);
    break;
  case MN_FUNCTION:
    c = (char *)malloc(strlen(c1)+strlen(c2)+3);
    sprintf(c, "%s(%s)", c1, c2);
    break;
  case MN_INDIRECT_FUNCTION:
    c = (char *)malloc(strlen(c1)+strlen(c2)+5);
    sprintf(c, "(%s)(%s)", c1, c2);
    break;
  case MN_EQ:
    c = (char *)malloc(strlen(c1)+strlen(c2)+5);
    sprintf(c, "(%s==%s)", c1, c2);
    break;
  case MN_NE:
    c = (char *)malloc(strlen(c1)+strlen(c2)+5);
    sprintf(c, "(%s!=%s)", c1, c2);
    break;
  case MN_GE:
    c = (char *)malloc(strlen(c1)+strlen(c2)+5);
    sprintf(c, "(%s>=%s)", c1, c2);
    break;
  case MN_GT:
    c = (char *)malloc(strlen(c1)+strlen(c2)+4);
    sprintf(c, "(%s>%s)", c1, c2);
    break;
  case MN_LE:
    c = (char *)malloc(strlen(c1)+strlen(c2)+5);
    sprintf(c, "(%s<=%s)", c1, c2);
    break;
  case MN_LT:
    c = (char *)malloc(strlen(c1)+strlen(c2)+4);
    sprintf(c, "(%s<%s)", c1, c2);
    break;
  case MN_ADD:
    c = (char *)malloc(strlen(c1)+strlen(c2)+4);
    sprintf(c, "(%s+%s)", c1, c2);
    break;
  case MN_SUB:
    c = (char *)malloc(strlen(c1)+strlen(c2)+4);
    sprintf(c, "(%s-%s)", c1, c2);
    break;
  case MN_MUL:
    c = (char *)malloc(strlen(c1)+strlen(c2)+4);
    sprintf(c, "(%s*%s)", c1, c2);
    break;
  case MN_DIV:
    c = (char *)malloc(strlen(c1)+strlen(c2)+4);
    sprintf(c, "(%s/%s)", c1, c2);
    break;
  case MN_MOD:
    c = (char *)malloc(strlen(c1)+strlen(c2)+4);
    sprintf(c, "(%s%%%s)", c1, c2);
    break;
  case MN_NEG:
    c = (char *)malloc(strlen(c1)+4);
    sprintf(c, "-(%s)", c1);
    break;
  case MN_OR:
    c = (char *)malloc(strlen(c1)+strlen(c2)+5);
    sprintf(c, "(%s||%s)", c1, c2);
    break;
  case MN_AND:
    c = (char *)malloc(strlen(c1)+strlen(c2)+5);
    sprintf(c, "(%s&&%s)", c1, c2);
    break;
  case MN_NOT:
    c = (char *)malloc(strlen(c1)+4);
    sprintf(c, "!(%s)", c1);
    break;
  case MN_CONDITION:
    c = (char *)malloc(strlen(c1)+strlen(c2)+strlen(c3)+9);
    sprintf(c, "(%s)?(%s):(%s)", c1, c2, c3);
    break;
  default:
    break;
  }

  if (c1)
    free(c1);
  if (c2)
    free(c2);
  if (c3)
    free(c3);

  e->upexpr = c;

  return e;
}

statement *make_statement(m)
     mnemonic m;
{
  statement *s;

  s = NEW(statement);
  s->next = s->jump = NULL;
  s->type = m;
  s->value = NULL;
  return s;
}

statement *append_statements(front, end)
     statement *front, *end;
{
  statement *temp;

  temp = front;

  if (!front)
    return end;

  while (temp->next)
    temp = temp->next;
  temp->next = end;

  return front;
}

#ifdef __STDC__
statement *make_line_stmt(string_list *local, string_list *global,
			  int len, char *form, ...)
#else
statement *make_line_stmt(local, global, len, form, va_alist)
     string_list *local;
     string_list *global;
     int len;
     char *form;
     va_dcl
#endif
{
  char *s;
  statement *rval;
  var *v;
  va_list list;

  s = (char *)malloc(len);
  VA_START(list, form);
  vsprintf(s, form, list);
  va_end(list);

  rval = make_statement(MN_LINE);

  v = newlist(4);
  v->list[0] = newnum(0);
  v->list[1] = newstring(s);
  v->list[2] = sl2var(local);
  v->list[3] = sl2var(global);

  rval->value = v;
  return rval;
}

/* this will destroy the original list */
var *sl2var(sl)
     string_list *sl;
{
  int i;
  string_list *temp;
  var *v;

  for (i = 0, temp = sl; temp; temp = temp->next, i++);

  v = newlist(i);

  for (i=0; sl; i++) {
    v->list[i] = newstring(sl->string);
    temp = sl;
    sl = sl->next;
    free(temp);
  }

  return v;
}

var *unparse(s)
     statement *s;
{
  int x;
  var *rval;
  statement *temp;

  x = 0;
  for (temp = s; temp; temp = temp->next)
    if (temp->type == MN_LINE)
      x = temp->value->list[0]->num;

  rval = newlist(x);

  for (temp = s; temp; temp = temp->next)
    if (temp->type == MN_LINE)
      rval->list[temp->value->list[0]->num-1]
	= newstring(mystrdup(temp->value->list[1]->string));

  return rval;
}
