/*
 * stalk/stuff.c
 * Copyright (c) 1993 Gerald Weiler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "include.h"
#include "stuff.h"
#include "stalk.h"
#include "functions.h"

void do_nothing(buf)
     buffer *buf;
{
  return;
}

void beep(buf)
     buffer *buf;
{
  if (bell_on)
    write(FD_OUT, "\007", 1);
}

void set_bell(bellstate)
     int bellstate;
{
  bell_on = bellstate;
}

/* taken from LambdaMOO, by Pavel Curtis */
int strindex(source, what, case_counts)
     char *source;
     char *what;
     int case_counts;
{
  char *s;
  int lwhat;

  lwhat = strlen(what);
  
  for (s = source; *s; s++) {
    if (!(case_counts ? strncmp(s, what, lwhat)
	  : mystrncasecmp(s, what, lwhat))) {
      return s - source + 1;
    }
  }
  return 0;
}

int strrindex(source, what, case_counts)
     char *source, *what;
     int case_counts;
{
  char *s;
  int lwhat;

  lwhat = strlen(what);
  
  for (s = source + strlen(source) - 1; s >= source; s--) {
    if (!(case_counts ? strncmp(s, what, lwhat)
	  : mystrncasecmp(s, what, lwhat))) {
      return s - source + 1;
    }
  }
  return 0;
}

char *strsub(source, what, with, case_counts)
     char *source, *what, *with;
     int case_counts;
{
  char *str;
  int lwhat = strlen(what);
  int x, slen, len;

  /*
   * allow it to expand in case everything is substituted with a bigger string,
   * but don't allow it to shrink, since we might not substitite anything at
   * all. Thanks to Hecubus and Omega for spotting this.
   */
  slen = strlen(source);
  len = MAX(slen, slen * strlen(with) / strlen(what));
  str = (char *)malloc(len+1);
  str[0] = '\0';

  while (*source) {
    if (! (case_counts ? strncmp(source, what, lwhat)
	   : mystrncasecmp(source, what, lwhat))) {
      strcat(str, with);
      source += lwhat;
    } else {
      x = strlen(str);
      str[x++] = source[0];
      str[x] = '\0';
      source++;
    }    
  }

  return str;
}

#ifdef __STDC__
void die(char *form, ...)
#else
void die(form, va_alist)
     char *form;
     va_dcl
#endif
{
  static int dielevel=0;
  va_list list;

  if (dielevel++) {
    /*
     * Ack! we've died before!
     * either that or something's been overwriting all memory, in which
     * case restoring the terminal state's meaningless, since it
     * might have been overwritten, too.
     * don't even bother doing anything.
     */
    abort();
  }

  (end_screen)();

  VA_START(list, form);

  /* construct the message via vfprintf */
  vfprintf(stdout, form, list);
  va_end(list);

  abort();
}

/* my_tsub will substitute param1 and param2 into a termcap control
 * string cstr. glibc has tparam, which works in place of this (probably
 * works better), but SunOS doesn't seem to have it - just tgoto (which
 * doesn't seem to work except for cursor movement strings).
 * this is based off of the parameter substitution guide in the SunOS
 * man pages and in the Nutshell Termcap & Terminfo manual */
char *my_tsub(cstr, param1, param2)
     char *cstr;
     int param1;
     int param2;
{
  char *rval, *tmp;
  int arg[3], current;

  arg[0] = param1;
  arg[1] = param2;
  current = 0;

  tmp = rval = malloc(3*strlen(cstr)); /* should be enough */
  while (*cstr) {
    if (*cstr == '%') {
      cstr++;
      switch(*cstr) {
      case 'd': /* send as decimal number */
	if (arg[current] > 99) {
	  *tmp++ = (char) arg[current]/100 + '0';
	  arg[current] %= 100;
	}
	if (arg[current] > 9) {
	  *tmp++ = (char) arg[current]/10 + '0';
	  arg[current] %= 10;
	}
	*tmp++ = (char) arg[current] + '0';
	current++;
	break;
      case '2': /* send as 2 digit decimal number */
	*tmp++ = (char) arg[current]/10 + '0';
	*tmp++ = (char) arg[current]%10 + '0';
	current++;
	break;
      case '3': /* send as 3 digit decimal number */
	*tmp++ = (char) arg[current]/100 + '0';
	*tmp++ = (char) (arg[current]/10)%10 + '0';
	*tmp++ = (char) arg[current]%10 + '0';
	current++;
	break;
      case '.': /* send as ASCII value */
	*tmp++ = (char) arg[current];
	current++;
	break;
      case '+': /* add ASCII value of next character from string
		 * and send ASCII character corresponding to the sum */
	*tmp++ = (char) arg[current] + *(++cstr);
	current++;
	break;
      case '%': /* send '%' */
	*tmp++ = '%';
	break;
      case 'r': /* reverse arguements */
	arg[2] = arg[0];
	arg[0] = arg[1];
	arg[1] = arg[2];
	break;
      case 'i': /* increment arguements by 1 */
	arg[0]++;
	arg[1]++;
	break;
      case '>': /* if ASCII value is greater than next character in string add
		 * the ASCII value of the character after that */
	if (arg[current] > *(++cstr))
	  arg[current] += *(cstr+1);
	++cstr;
	break;
      case 'n': /* exclusive OR the values with 01400 before continuing */
	arg[0] ^= 0140;
	arg[1] ^= 0140;
	break;
      case 'B': /* change next value to binary-coded-decimal.
		 * ((16*(x/10)) + (rem(x/10))) */
	arg[current] += 6*(arg[current]/10);
	break;
      case 'D': /* the next value is reverse coded (x-2*(rem(x/16))) */
	arg[current] -= 2*(arg[current]%16);
	break;
      }
      cstr++;
      /**/
    } else
      *tmp++ = *cstr++;
  }

  *tmp = '\0';
  return rval;
}

char *mystrdup(s)
     char *s;
{
  char *r, *t;

  t = r = malloc(strlen(s)+2);
  while (*s)
    *t++ = *s++;
  *t = '\0';
  return r;
}

int mystrcasecmp(s1, s2)
     char *s1;
     char *s2;
{
  int rval;

  while (!(rval = tolower(*s1) - tolower(*s2)) && *s1++)
    s2++;

  return rval;
}

int mystrncasecmp(s1, s2, n)
     char *s1;
     char *s2;
     int n;
{
  int rval;

  rval = 0;

  while (n-- && !(rval = tolower(*s1) - tolower(*s2)) && *s1++)
    s2++;

  return rval;
}

/*
 * ptym_open and ptys_open based on similar functions in 'Advanced
 * Programming in the UNIX Environment' by W. Richard Stevens. A
 * wonderful book.
 */

/*
 * find and open next pty master
 */
int ptym_open(name)
     char *name;
{
#ifdef SYSVTTY
  char *s;
  int mfd;

  strcpy(name, "/dev/ptymx");
  if ((mfd = open(name, O_RDWR)) < 0)
    return -1;

  if (grantpt(mfd) <0) {
    close(mfd);
    return -1;
  }

  if (unlockpt(mfd) <0) {
    close(mfd);
    return -1;
  }

  if ((s = ptsname(mfd)) == NULL) {
    close(mfd);
    return -1;
  }

  strcpy(name, s);
  return mfd;
#else
  /* use BSDTTY */

  int mfd;
  char *s1, *s2;

  strcpy(name, "/dev/ptyXY");
  for (s1 = "pqrstuvwxyzPQRST"; *s1; s1++) {
    name[8] = *s1;
    for (s2 = "0123456789abcdef"; *s2; s2++) {
      name[9] = *s2;

      /* try to find a master */
      if ((mfd = open(name, O_RDWR)) < 0) {
	if (errno == ENOENT)
	  /* out of ptys. what a pity (you knew that was coming...) */
	  return -1;
	else
	  continue;
      }

      name[5] = 't'; /* change 'pty' to 'tty' */
      return mfd;
    }
  }

  return -1;
#endif /* SYSVTTY */
}

/*
 * open pty slave associated with name
 */
int ptys_open(mfd, name)
     int mfd;
     char *name;
{
#ifdef SYSVTTY
  int sfd;

  if ((sfd = open(name, O_RDWR)) < 0) {
    close(mfd);
    return -1;
  }

  if ((ioctl(sfd, I_PUSH, "ptem") < 0) ||
      (ioctl(sfd, I_PUSH, "ldterm") < 0) ||
      (ioctl(sfd, I_PUSH, "ttcompat") < 0) ||) {
    close(mfd);
    close(sfd);
    return -1;
  }

  return sfd;
#else
  /* use BSDTTY */

  int sfd, gid;
  struct group *gptr;

  gptr = getgrnam("tty");
  if (gptr)
    gid = gptr->gr_gid;
  else
    gid = -1;

  chown(name, getuid(), gid);
  chmod(name, S_IRUSR | S_IWUSR | S_IWGRP);

  if ((sfd = open(name, O_RDWR)) < 0) {
    close(mfd);
    return -1;
  }

  return sfd;
#endif /* SYSVTTY */
}

int all_space(s)
     char *s;
{
  if (!s)
    return 1;

  while (*s) {
    if (!isspace(s[0]))
      return 0;
    s++;
  }

  return 1;
}

char *expand_filename(filename)
     char *filename;
{
  char *temp, *username;
  int i;
  struct passwd *pwd;

  if (!filename)
    return NULL;

  if (filename[0] == '~') {
    if (filename[1] == '/') {
      pwd = getpwuid(geteuid());
      if (!pwd)
	return mystrdup(filename);
      temp = (char *)malloc(strlen(pwd->pw_dir)+strlen(filename));
      sprintf(temp, "%s/%s", pwd->pw_dir, filename+2);
      return temp;
    } else {
      i=1;
      while (filename[i] && (filename[i] != '/'))
	i++;

      username = (char *)malloc(i);

      i=1;
      while (filename[i] && (filename[i] != '/')) {
	username[i-1] = filename[i];
	i++;
      }
      username[i-1] = '\0';

      pwd = getpwnam(username);
      if (!pwd) {
	free(username);
	return filename;
      }

      temp = (char *)malloc(strlen(pwd->pw_dir)+strlen(filename+i)+2);
      sprintf(temp, "%s/%s", pwd->pw_dir, filename+i);
      free(username);
      return temp;
    }
  } else
    return mystrdup(filename);
}
